# coding=utf-8

"""
    vcfy.cli
    ~~~~~~~~

    This module implements command-line interface.

    :copyright: (c) 2018 by Ali Ghaffaari.
    :license: MIT, see LICENSE for more details.
"""

import click

from . import model


@click.command()
@click.argument('reference', type=click.File('r'))
@click.option('-o', '--output', default="-", type=click.File('w'),
              help="Write to this file instead of standard output.")
@click.option('-m', '--mutation-rate', type=float, required=True,
              help="Base mutation rate.")
@click.option('-r', '--region', type=str, default=None,
              help="Region ID (default=first region in the reference)")
@click.option('-l', '--low', type=int, default=None,
              help="Range lower bound (default=first locus in the region)")
@click.option('-h', '--high', type=int, default=None,
              help="Range upper bound (default=last locus in the region)")
def cli(**kwargs):
    """Generate VCF file with simulated variants in specified range [low, high)
    for the given region of the reference genome. In case that the region is not
    specified the first region is used. If no range is provided, it is assumed
    that the variants are scattered throughout the region.

    For more information, consult with the README file.
    """
    model.generate_vcf(
        ref=kwargs.pop('reference'),
        vcf_out=kwargs.pop('output'),
        region_id=kwargs.pop('region'),
        mrate=kwargs.pop('mutation_rate'),
        **kwargs)



# coding=utf-8

"""
    vcfy.model
    ~~~~~~~~~~

    This module implements the model for generating variants based on the given
    parameters.

    :copyright: (c) 2018 by Ali Ghaffaari.
    :license: MIT, see LICENSE for more details.
"""

from numpy import random, arange
import vcf

from . import util


def rnd_sv(locus, seq):
    """Get random structural variant for the given `locus` in the `seq`.

    Parameters:
        locus : int
            The position in the sequence for which a random SV is generated.
        seq : str
            The reference DNA sequence.

    Return:
        A tuple indicating the REF and ALT alleles.
    """
    ref = seq[locus-1].upper()
    if ref not in util.BASES:
        raise RuntimeError("invalid base character")

    return ref, random.choice([b for b in util.BASES if b != ref])


def simulate(region, mrate, low=None, high=None):
    """Simulate variants for the given region in the one-based range
    [low, high) using the probability model defined by mass function `pmf`.

    Parameters:
        region : Bio.SeqRecord.SeqRecord
            The BioPython's `SeqRecord`-like object of the region containing the
            region ID and its sequence.
        mrate : float
            Base mutation rate.
        low : int, optional
            The lower bound of the range in which the variants are simulated. It
            is assumed to be 1, if not provided.
        high : int, optional
            If provided, it is one above the upper bound of the range; otherwise
            it is set to the length of the region sequence.
    Return:
        A dictionary containing values for these VCF fields: POS, ID, REF, ALT,
        QUAL, and FILTER.
    """
    low = 1 if low is None else max(1, low)
    high = len(region.seq) + 1 if high is None else min(high, len(region.seq)+1)

    for locus in arange(low, high):
        if random.choice([True, False], p=[mrate, 1-mrate]):
            try:
                ref, alt = rnd_sv(locus, region.seq)
            except RuntimeError as err:
                util.warn(err)
                continue
            yield dict(POS=locus,
                       ID=util.VCF_MISSING_VALUE,
                       REF=ref,
                       ALT=alt,
                       QUAL=util.VCF_MISSING_VALUE,
                       FILTER=util.VCF_MISSING_VALUE)


def generate_vcf(ref, vcf_out, region_id=None, **sim_params):
    """Generate simulated VCF file.

    Parameters:
        ref : str or opened file
            File path or object of the reference genome FASTA file.
        vcf_out : str or opened file
            File path or writable file object corresponding to output VCF file.
        region_id : str
            The ID of the region for which the variants are simulated.
        sim_params: keyword paramters
            The parameters required for simulation passed directly to the
            `simulate` function.
    """
    if region_id is None:
        region = next(util.filter_regions(ref, n=1))
    else:
        region = next(util.filter_regions(ref, include=[region_id]))

    template = vcf.Reader(util.make_template(ref, region, **sim_params))

    if isinstance(vcf_out, str):
        writer = vcf.Writer(open(vcf_out, 'w'), template)
    else:
        writer = vcf.Writer(vcf_out, template)

    tmpl_record = next(template)
    for svar in simulate(region, **sim_params):
        record = util.update_record(tmpl_record, CHROM=region.id, **svar)
        writer.write_record(record)



# coding=utf-8

"""
    vcfy
    ~~~~

    Generate a VCF file with random variants using the given probability model.

    :copyright: (c) 2018 by Ali Ghaffaari.
    :license: MIT, see LICENSE for more details.
"""



# coding=utf-8

"""
    vcfy.release
    ~~~~~~~~~~~~

    Module containing package release information.

    :copyright: (c) 2018 by Ali Ghaffaari.
    :license: MIT, see LICENSE for more details.
"""

# CONSTANTS ###################################################################
# Development statuses:
DS_PLANNING = 1
DS_PREALPHA = 2
DS_ALPHA = 3
DS_BETA = 4
DS_STABLE = 5
DS_MATURE = 6
DS_INACTIVE = 7
DS_STRING = {
    DS_PLANNING: 'Development Status :: 1 - Planning',
    DS_PREALPHA: 'Development Status :: 2 - Pre-Alpha',
    DS_ALPHA: 'Development Status :: 3 - Alpha',
    DS_BETA: 'Development Status :: 4 - Beta',
    DS_STABLE: 'Development Status :: 5 - Production/Stable',
    DS_MATURE: 'Development Status :: 6 - Mature',
    DS_INACTIVE: 'Development Status :: 7 - Inactive'
}
###############################################################################

# Package release information.
__title__ = 'vcfy'
__description__ = 'Generate VCF file with random variants from reference genome'
__author__ = 'Ali Ghaffaari'
__email__ = 'ali.ghaffaari@mpi-inf.mpg.de'
__license__ = 'MIT'

# Release
__version__ = '0.0.4'
__status__ = DS_PREALPHA

# Package data
__package_data__ = {__title__: ['resources/templates/template.vcf']}

# PyPI-related information
__keywords__ = 'vcf genome genomics variation DNA sequence ngs'
__classifiers__ = [
    # Development status
    DS_STRING[__status__],

    # Environment
    'Environment :: Console',

    # License
    'License :: OSI Approved :: MIT License',

    # Supported Python versions.
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',

    # Intended Audience and Topic
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
]
__requires__ = [
    'PyVCF>=0.6.8',
    'biopython>=1.71',
    'numpy>=1.14.3',
    'click>=6.7',
]
__tests_require__ = []
__extras_require__ = {
    'test': ['nose>=1.0', 'coverage'],
}
__setup_requires__ = ['nose>=1.0', 'coverage']
__entry_points__ = '''
[console_scripts]
vcfy=vcfy.cli:cli
'''



# coding=utf-8

"""
    vcfy.resources
    ~~~~~~~~~~~~~~

    This subpackage contains required resources.

    :copyright: (c) 2018 by Ali Ghaffaari.
    :license: MIT, see LICENSE for more details.
"""



# coding=utf-8

"""
    vcfy.util
    ~~~~~~~~~

    The utility module for helper classes and functions.

    :copyright: (c) by 2018 by Ali Ghaffaari.
    :license: MIT, see LICENSE for more details.
"""

from __future__ import print_function
import os
import sys
import copy
import datetime
import tempfile
import pkg_resources as pkg_res

from Bio import SeqIO

from . import release


BASES = ['A', 'C', 'G', 'T']
VCF_MISSING_VALUE = '.'
VCF_TEMPLATE = 'resources/templates/template.vcf'


def warn(msg):
    """Print a warning message."""
    print("[{pkg}] WARNING: {msg}.".format(pkg=release.__title__, msg=msg),
          file=sys.stderr)


def make_template(ref, region, **params):
    """Fill wildcards in the templated VCF file and create a ready-to-use
    template.

    Parameters:
        ref : str or opened file
            File path or object of the reference genome FASTA file.
        region : SeqIORecord
            The region FASTA record.

    Return:
        The file path of newly created template VCF file.
    """
    wildcards = dict()
    wildcards['date'] = datetime.datetime.today().strftime('%Y%m%d')
    wildcards['version'] = release.__version__
    wildcards['reference'] = os.path.realpath(ref.name if hasattr(ref, 'name')
                                              else ref)
    wildcards['contigs'] = ('<ID=' + region.id +
                            ',length=' + str(len(region.seq)) + '>')
    opts = ['vcfy']
    opts.append('-m ' + str(params['mrate']))
    opts.append("-r '" + region.id + "'")
    opts.append('-l ' + str(params['low']) if params['low'] else '')
    opts.append('-h ' + str(params['high']) if params['high'] else '')
    wildcards['cmd'] = ' '.join(o for o in opts if o)
    template = tempfile.NamedTemporaryFile(mode='w+t')
    for line in open(pkg_res.resource_filename(__name__, VCF_TEMPLATE), 'r'):
        template.write(line.format(**wildcards))
    template.seek(0)
    return template


def filter_regions(fasta_file, **kwargs):
    """Yield the SeqIORecord of the regions in the FASTA file filtered based on
    the given parameters. In case of no filter parameters are provided, all
    regions is reported.

    Parameters:
        fasta_file : str or opened file
            File path or object of the input FASTA file.
        include : list of str, optional
            List of region IDs to be *included* to the return dictionary (will
            be overriden by `exclude` if there is any conflict between these two
            lists).
        exclude : list of str, optional
            List of region IDs to be *excluded* from the return dictionary.
        n : int, optional
            If provided, the first `n` number of regions are returned; otherwise
            all regions are be returned if neither `include` nor `exclude` is
            provided. If this parameter is set to negative values, all regions
            are returned.

    Return:
        A dictionary containing DNA sequence of each region with its ID as key.
    """
    include = kwargs.pop('include', [])
    exclude = kwargs.pop('exclude', [])
    num = kwargs.pop('n', -1) if not include and not exclude else -1

    fasta = SeqIO.parse(fasta_file, 'fasta')
    for region in fasta:
        if num == 0:
            break
        if region.id in exclude or (include and region.id not in include):
            continue
        yield region
        num -= 1


def update_record(old, **kwargs):
    """Get a copy of the given VCF record with updated fields with new values.

    Parameters:
        old : vcf.model._Record
            The record to be update.
        kwargs : dict
            The new values for each field provided by keyword parameters.

    Return:
        The updated copy of the record.
    """
    record = copy.deepcopy(old)
    for key, value in kwargs.items():
        setattr(record, key, value)
    return record



# coding=utf-8

"""
    setup.py
    ~~~~~~~~

    Setup script.

    :copyright: (c) 2018 by Ali Ghaffaari.
    :license: MIT, see LICENSE for more details.
"""

import os
import codecs
from setuptools import setup, find_packages


_HERE = os.path.abspath(os.path.dirname(__file__))

# Get package release information.
_PACKAGE_NAME = "vcfy"
_RELEASE_INFO = {}
with open(os.path.join(_HERE, _PACKAGE_NAME, 'release.py')) as release_file:
    exec(release_file.read(), _RELEASE_INFO)  # pylint: disable=exec-used

# Get the long description from the README file.
with codecs.open(os.path.join(_HERE, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

_PYPINAME = "vcfy"
_PACKAGES = find_packages(exclude=['test'])
_GITHUB_BASE = "https://github.com/cartoonist/"
_VCS_URL = _GITHUB_BASE + _PYPINAME
_TAR_URL = _VCS_URL + "/tarball/" + _RELEASE_INFO['__version__']

setup(
    name=_PYPINAME,
    packages=_PACKAGES,
    package_data=_RELEASE_INFO['__package_data__'],
    version=_RELEASE_INFO['__version__'],
    description=_RELEASE_INFO['__description__'],
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author=_RELEASE_INFO['__author__'],
    author_email=_RELEASE_INFO['__email__'],
    license=_RELEASE_INFO['__license__'],
    url=_VCS_URL,
    download_url=_TAR_URL,
    keywords=_RELEASE_INFO['__keywords__'],
    classifiers=_RELEASE_INFO['__classifiers__'],
    install_requires=_RELEASE_INFO['__requires__'],
    tests_require=_RELEASE_INFO['__tests_require__'],
    extras_require=_RELEASE_INFO['__extras_require__'],
    setup_requires=_RELEASE_INFO['__setup_requires__'],
    entry_points=_RELEASE_INFO['__entry_points__'],
)
